from __future__ import division
import matplotlib
matplotlib.use('agg')
import json
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.colors as cl
import pickle
import pprint
import time
import pandas
import datetime
import calendar
from itertools import groupby
from collections import defaultdict
import itertools
import random,os,glob
from shapely.geometry import Point

control=pandas.read_csv('user_type.csv','r', delimiter=',', names=None, squeeze=False, skipinitialspace=True,usecols=['id','iata','is_worker'])
#control is a preprocessed dataset identifying workers and passengers of LHR airport
#id is the anonymized id of the user

control=control[control.iata=='LHR'] #we want only heathrow

cc=set(control['id']) #get the set of unique ids
control.set_index('id', inplace=True)

dfs=[]
path = 'newdata/'

nms=['0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'] #original dataset was cut into smaller files

for i in nms:
    df=pandas.read_csv(path+'sort_'+i+'_filt.csv','r', delimiter=',', header=None, skiprows=1, names=None, index_col=None, squeeze=False, skipinitialspace=True)
    print(i)
    dfs.append(df)

data=pandas.concat(dfs)

#0 was the position of the user id
#1 time
#3,4 are latitude and longitude coordinates

print ('n users',len(data.groupby(0))) #this gathers the dataset in groups of users, hence you get for each user a set of points in space

data = data.sort_values(1) #sort by time

data = data.groupby(0).filter(lambda x: len(x)>10) #drop those users who have less than 10 points, not representative


def oneyear(x): #security check to avoid using sparse datapoints from other years
    t=x
    tt1=int(time.strftime('%Y', time.gmtime(float(t)))) #week number
    if tt1==2017:
        return True
    else: return False

data = data[data[1].apply(lambda x: oneyear(x))] #this checks the time column of all points and filters out spurious data points from different years

weeksecs=604800 #one week seconds

data.set_index(0, inplace=True)

daysecs=86400 #one day seconds

data['nome']=data.index #I create a new column copying users ids
data['newtime']=data[1] #same for time, this is where I shift trajectories from other days and assign them a new user id

supname=dict() #double key dictionary, 1st one is the user, 2nd one is the day
chosenday=107 #chosenday = choose the reference day (int, day of the year), in this case the first day of the dataset (if not, change the code below)

def supersample(x):
        
    ti=x[1]
    tti=int(time.strftime('%j', time.gmtime(float(ti)))) #tti is the day of the data point

    conn=supname.get(x['nome'],'error') #this gives 'error' if the user is not in the dictionary yet
    if conn=='error': #if user is not in the dict yet
        supname[x['nome']]={} #for each user id creates a key in the dictionary, then I will make a second key for the days
        
    conn2=supname[x['nome']].get(tti,'error') #this gives 'error' if the user does not have any point in "tti" day
    
    if conn2=='error': #if this user did not appear before during this day
        pp=list(str(x['nome'])) #string comprehension
        g=int(random.random()*20) #i generate a random number from 1 to 10
        supname[x['nome']][tti]=str("".join(pp[0:-g])) #I change its name by cutting its old name (a long encrypted string) to make it "g"-length shorter
        
    if tti==chosenday: #if it is the first day of the dataset (chosenday, up to you) everything can be left as it is
        x['nome']=x['nome']
        x['newtime']=int(x[1])

    else:
        daydiff=tti-chosenday #n of days from chosenday
        tnew = ti-(daysecs*daydiff) #n of seconds from chosenday

        x['nome']=supname[x['nome']][tti] #get the new name I already created for this user for day 'tti'
        x['newtime']=int(tnew) #shift back the trajectory points to the day I want (chosenday)

    return x

data=data.apply(supersample,axis=1) #this is the function that makes the supersampling

print (len(data)) #check the length of the dataset now

print ('n users super',len(data.groupby('nome'))) #group the dataset by the new names, hence count the new users as well

tmax=np.max(data['newtime']) #I define the minimum time overall
tmin=np.min(data['newtime']) #and maximum, in seconds

tmaxn=int(time.strftime('%W', time.gmtime(float(tmax))))
tminn=int(time.strftime('%W', time.gmtime(float(tmin))))

T=(tmax-tmin)/900 +1  #900 seconds = 15 minutes
print ('number of timecells')  #I find the number of timeslots of 15min in the whole database time window

lat, lon =  51.470020,-0.454295 #heathrow coordinates

from pyproj import Proj
myProj = Proj("+proj=utm +zone=30U, +north +ellps=WGS72 +units=m +no_defs") #define the projection

xa, ya = myProj(lon,lat)  #project lat lon to meters
print ('xa, ya')
print (xa,ya)

import codecs
import heapq
import glob
import math
import numpy as np
import shapefile
import matplotlib.cm as cm
import matplotlib.patches as patches
from matplotlib.patches import Polygon, Circle, Rectangle
from matplotlib.collections import PatchCollection
from shapely.geometry import Point
from numpy import array

deltax=10
sf = shapefile.Reader('/home/mattia/Tools/Airports/script/New/uk/LHR10m.shp') #read the file of heathrow airport 10mx10m grid
#I made it from qGis, get UK shapefile, focus on heathrow coordinates, project it to the Mercator projection, get the vectorial grid of 10x10m

shapes  = sf.shapes() #cells of the shapefile
Nshp    = len(shapes) #number of cells

shapes  = sf.shapes() #cells of the shapefile
Nshp    = len(shapes) #number of cells

centros=dict()
xy=dict()
#recs    = sf.records() #read properties of shp file
#fields = sf.fields

pkl_file = open('/home/mattia/Tools/Airports/script/New/LHR10mcentros.pkl','rb')
centros = pickle.load(pkl_file)
pkl_file.close()

for nshp in xrange(Nshp): #as long as there are cells

    ww=0
    ptchs   = []
    pts     = array(shapes[nshp].points) #list of geographic points of every cell
    prt     = shapes[nshp].parts
    par     = list(prt) + [pts.shape[0]]

    centroid=[pts[0][0]+(pts[2][0]-pts[0][0])/2,pts[2][1]+(pts[0][1]-pts[2][1])/2]

    centros[nshp]=centroid #I define the controids of each cell in the grid

print ("len centros")
print (len(centros))

del sf
#delete unnecessary heavy variables

X=[]
Y=[]

for i in range(len(centros)):
    X.append(centros[i][0])
    Y.append(centros[i][1])
    
#in the next lines I get the Xmin,Ymin,Xmax,Ymax of the grid
x0=shapes[0].points[0][0] #0left-top 1right-top 2right-bottom 3left-bottom
y0=shapes[0].points[0][1]
xmax=shapes[-1].points[1][0] 
ymax=shapes[-1].points[2][1]
#deltax is 10m, the spatial scale I want
Lx=float("%.2f" % (xmax-x0))/deltax #here I get how many cells I have along the x axis and the y axis
Ly=float("%.2f" % (y0-ymax))/deltax

nx=int(Lx)
ny=int(Ly)

def projection(x):
    x0,y0=myProj(x[3],x[2])
    return x0,y0

data[6] = data.apply(projection,axis=1) #project lat lon coordinates of users to mercator as well (6 is now the column of projected data points coordinates)

def collocate(x):
    xi=x[0]
    yi=x[1]
    xx=int((xi-x0)/deltax) #num columns
    yy=int((y0-yi)/deltax) #num rows

    if xx<0 or yy<0 or xx>nx or yy>ny:
        return 0
    else:    
        a=np.ravel_multi_index((yy,xx), dims=(ny,nx),mode='clip') #(rows,cols)
        return a
                    
data['cell'] = data[6].apply(collocate) #place trajectories points into cells


data=data[data.cell!=0] #remove points that fall out of the grid

data = data[['nome','newtime','cell',6]] #cut the database to the desired columns only


from collections import Counter

tmax=np.max(data['newtime'])
tmin=np.min(data['newtime'])

data=data.sort_values(['nome','newtime'])

print ('after the grid')
print (len(data.groupby('nome'))) #check how many users are left after the grid placement filter

import itertools

def timecell(x):
    a=int((x-tmin)/900)
    return a

data['timecell']=data['newtime'].apply(timecell) #make a new column with the timeslot index instead of seconds for each trajectory point

data = data.drop_duplicates(['nome','cell','timecell']) #remove duplicates of space and time, we already know the user was in cell A at time B

def identity2(x):
    xx=x.name
    if control['is_worker'].loc[xx]==True:
        return 'w'
    else:
        return 't'
    
data['typez'] = data.apply(identity2,axis=1) #make a new column from the control databaset, to divide travelers from workers

tipo = dict(zip(data.nome, data.typez)) #make the dictionary with user id and category (travelers, worker)

data2=data[data.typez=='t'] #data2 is the dataset of travelers only, now i want to classify them into arrivals, connecting, departures

data2 = data2[['nome','cell','newtime',6,'timecell']] #cut it to the desired columns

first=[]
last=[]

#the next coordinates are the bounding boxes of terminals and landing strips, I need them to classify travelers into departing, arrivals exc

xt2min=677300  #t2
xt2max=677800
yt2min=5705000
yt2max=5705750

xt3min=675800  #t3
xt3max=676400
yt3min=5704800
yt3max=5705750

xt5min=674300  #t5
xt5max=675750
yt5min=5704700
yt5max=5705750

xt4min=677000   #t4
xt4max=677700
yt4min=5703700
yt4max=5704500

xt1min=675800   #gates2e3
xt1max=677800
yt1min=5704800
yt1max=5705200

xt6min=676500   #central station
xt6max=677000 
yt6min=5705250
yt6max=5705500

xt7min=678800  #train station
xt7max=679300
yt7min=5704500
yt7max=5705000

xtp1min=674400 #landing strip 1
xtp1max=678300
ytp1min=5704500
ytp1max=5704800

xtp2min=674300 #landing strip 2
xtp2max=678300
ytp2min=5705700
ytp2max=5706200

xt8min=675000  #facility
xt8max=676800
yt8min=5703500
yt8max=5704500

xt9min=674000 #t5 metro
xt9max=674300
yt9min=5704600
yt9max=5705700


times=[]
tipo2=dict()

trajboh=[]
traj1=[]

#the next for loop is the classification algorithm of travelers, it uses semantics as well checking the order of visit of specific places

for a,group in data2.groupby('nome'): #group trajectories with new names, for each new user ->
    
    d=list(group['cell'].get_values())

    gg=group.sort_values('newtime')
    first.append(gg['cell'].values[0])
    last.append(gg['cell'].values[-1])
    lapse=gg['newtime'].values[-1]-gg['newtime'].values[0]
    times.append(lapse)
    
    x0=gg[6].values[0][0]   
    y0=gg[6].values[0][1]
    xf=gg[6].values[-1][0]    
    yf=gg[6].values[-1][1] 
    nome=gg['nome'].values[0]
    
    if  any([all([xt1min<x0<xt1max,yt1min<y0<yt1max]),all([xt2min<x0<xt2max,yt2min<y0<yt2max]),all([xt3min<x0<xt3max,yt3min<y0<yt3max]),all([xt4min<x0<xt4max,yt4min<y0<yt4max]),all([xt5min<x0<xt5max,yt5min<y0<yt5max])]):
        #if they appear at the gates

        if any([all([xt1min<xf<xt1max,yt1min<yf<yt1max]),all([xt2min<xf<xt2max,yt2min<yf<yt2max]),all([xt3min<xf<xt3max,yt3min<yf<yt3max]),all([xt4min<xf<xt4max,yt4min<yf<yt4max]),all([xt5min<xf<xt5max,yt5min<yf<yt5max])]):
            tipo2[nome]='c' #if they disappear at the gates
        
        else: 
            tipo2[nome]='a'
            
    elif any([all([xt9min<x0<xt9max,yt9min<y0<yt9max]),all([xt7min<x0<xt7max,yt7min<y0<yt7max]),all([xt6min<x0<xt6max,yt6min<y0<yt6max])]):
        tipo2[nome]='d'
        
    elif any([all([xt1min<xf<xt1max,yt1min<yf<yt1max]),all([xt2min<xf<xt2max,yt2min<yf<yt2max]),all([xt3min<xf<xt3max,yt3min<yf<yt3max]),all([xt4min<xf<xt4max,yt4min<yf<yt4max]),all([xt5min<xf<xt5max,yt5min<yf<yt5max])]):
        tipo2[nome]='d'
    
    elif any([all([xt9min<xf<xt9max,yt9min<yf<yt9max]),all([xt7min<xf<xt7max,yt7min<yf<yt7max]),all([xt6min<xf<xt6max,yt6min<yf<yt6max])]):
        tipo2[nome]='a'
=
    elif any([all([xt8min<x0<xt8max,yt8min<y0<yt8max]),all([xt8min<xf<xt8max,yt8min<yf<yt8max])]):
        tipo2[nome]='w'

        
    else:
        tipo2[nome]='w2'
        traj=[]  #letters sequences
        trajxy=[] #trajectory points
        for i in range(len(gg)):
            xb=gg[6].values[i][0]    
            yb=gg[6].values[i][1] 
            trajxy.append((xb,yb))
            if all([xt1min<xb<xt1max,yt1min<yb<yt1max]):
                traj.append('A')
            elif all([xt2min<xb<xt2max,yt2min<yb<yt2max]):
                traj.append('B')
            elif all([xt3min<xb<xt3max,yt3min<yb<yt3max]):
                traj.append('C')
            elif all([xt4min<xb<xt4max,yt4min<yb<yt4max]):
                traj.append('D')
            elif all([xt5min<xb<xt5max,yt5min<yb<yt5max]):
                traj.append('E')
            elif all([xtp1min<xb<xtp1max,ytp1min<yb<ytp1max]):
                traj.append('P')
            elif all([xtp2min<xb<xtp2max,ytp2min<yb<ytp2max]):
                traj.append('P')
            elif all([xt8min<xb<xt8max,yt8min<yb<yt8max]):
                tipo2[nome]='w'
            
        c=Counter(traj)
        
        if len(c.keys())>3:
            tipo2[nome]='w'
        
        if c.get('P','error')!='error':
            if traj[0]!='P' and traj[-1]!='P':
                tipo2[nome]='w'
                
        if len(traj)>1:
            if traj[0]=='P' and traj[-1]!='P':
                tipo2[nome]='a'
            
            if traj[0]!='P' and traj[-1]=='P':
                tipo2[nome]='d'
        
        if len(c.keys())==1:
            if any([c.keys()=='A',c.keys()=='B',c.keys()=='C',c.keys()=='D',c.keys()=='E']): #users with only one semanting entry, not enough info, to discard
                tipo2[nome]='w2'
            

print (Counter(tipo2.values())) #check how many users you have in each category

del data2

tipo3=dict() #define a new dictionary where now I have workers, departing, arrivals and connecting travelers
for i in data['nome']:
    if tipo[i]=='w':
        tipo3[i]='w'
    else: tipo3[i]=tipo2[i]

def filtro(x):
    if tipo3[x.name]=='w2':
        return False
    else:
        return True
    
data=data.groupby('nome').filter(lambda x: filtro(x)) #w2 are bad trajectories users, I remove them

print (len(data.groupby('nome'))) #check how many are left

##################### completing trajectories

data = data[['nome','cell','timecell']]

## !! you can change which threshold you want for your optimal path in terms of number of visits per cell

pkl_file = open('cellssp30.pkl','rb') #here I load the data driven shortest path connecting every pair of cells with at least 30 presences in the dataset (already computed)
p = pickle.load(pkl_file)
pkl_file.close()

pkl_file = open('neighsSP.pkl','rb') #network of cardinal neighbors of each cell (up,down,left,right)
neighs = pickle.load(pkl_file)
pkl_file.close()


"""#the code to make the neighbors network, comment if already computed
X=[]
Y=[]

for i in range(len(centros)):
    X.append(centros[i][0])
    Y.append(centros[i][1])
    
ny=int((max(Y)-min(Y))/10)+1
nx=int((max(X)-min(X))/10)+1

print ('n cells')
print (nx,ny)
print (nx*ny,len(centros))


cells=range(0,len(centros))
neighs=defaultdict(list)

for i in cells:
    idy, idx = np.unravel_index(i, dims=(ny, nx)) #idx = n column, idy= n row
    neighx = (idx-1, idx+1, idx, idx)
    neighy = (idy, idy, idy-1, idy+1) #idy-1 means the upper neighbor
    neig=zip(neighy,neighx)
    
    for k in neig:
        #dictionary of neighbors
        neighs[i].append(np.ravel_multi_index(k, dims=(ny,nx),mode='clip'))

output = open('neighsSP.pkl','wb')
pickle.dump(neighs, output)
output.close()"""


datas = pandas.DataFrame(columns=['nome','cell','timecell']) #new points to be added to complete trajectories

def complete(x):
    df = pandas.DataFrame(columns=['nome','cell','timecell']) #new points to be added later

    nome=x['nome'].values[0]
    
    for i in range(len(x)):
        if i==len(x)-1: #when you get to the last point, stop the loop
            break
        else:

            c0=x['cell'].values[i] #check cells and time slots of each point and the next one
            t0=x['timecell'].values[i]
            cf=x['cell'].values[i+1]
            tf=x['timecell'].values[i+1]
            
            if cf not in neighs[c0]: #if cf is not an adjacent cell for c0
                if cf!=c0:
                    cond=p.get(c0,'e') #is c0 in the shortest path dictionary
                    if cond!='e':
                        cont=p[c0].get(cf,'e') #is cf an ending of a shortest path from c0?
                        if cont!='e': #if yes
                            lung=len(p[c0][cf]) #check how long is the path from c0 to cf
                            celdas=p[c0][cf]
                            tempi=np.linspace(t0,tf,lung,dtype=int) #check in how many time slots we have to divide the path
                            for a in range(lung):
                                
                                celda=celdas[a]
                                timece=tempi[a] #the user can appear in the same time slot in two different cells, it is ok, always has been
                                df.loc[len(df)]=[nome,celda,timece] #add new cell from the path and new time slot from the division
    #print (len(df))
    return df 
                                    
datas=data.groupby('nome').apply(complete) #now I complete the trajectories

data3=pandas.concat([data,datas]) #join the two datasets

data=data3

del data3
del datas

data = data.sort_values(['nome','timecell'])

print (len(data.groupby('nome')), 'tot users')
print (len(data), 'tot data points')
##################### completing trajectories, end


data = data.drop_duplicates(['nome','cell','timecell'])

print (len(data.groupby('nome')), 'tot users')
print (len(data), 'tot data points')

names=data['nome'].values[:]

d = {ni: indi for indi, ni in enumerate(set(names))} #transform users id to numbers

def cambio(x):
    return d[x]

data['num']=data['nome'].apply(cambio)

tipo4=dict()
for i in data['nome']:
    x=d[i]
    tipo4[x]=tipo3[i] #define the dictionary of categories, but now with users numerical identities instead of strings

output = open('tipo4SP30.pkl','wb')
pickle.dump(tipo4, output)
output.close()

# !! change these names (50, 40, exc) before running a new code

contact=np.zeros((4, 4)) #this is just a matrix I need to check the contacts among all categoriess
#here I define the contact network for each time slot, in each time slot I have a list of edges based on the reconstructed trajectories
edgelist=defaultdict(list) #dictionary: each time slots->active links

for a,group in data.groupby(['cell','timecell']): #group by same cells and same time slots
    
    dd=list(group['num'].get_values())
    if len(dd)>0: #to be sure i'm not considering a time space where i find only one user
        c = list(itertools.product(dd,dd)) #get all combinations of users
        t=group['timecell'].get_values()[0]
        
        cell=group['cell'].get_values()[0]
        for ll in c:
            if ll[0]!=ll[1]:
                edgelist[t].append((ll[0],ll[1],cell)) #generate links everytime a couple of users is found
                
from collections import Counter

tstart=0 #first time slot
t=tstart

tend=96
edge=[]
nodes = []
    
while t<tend:
    print (t)
    net = open('1day/2network_t'+str(t)+'.dat','w') #save the contact network, one file for each time slot
    
    links=edgelist[t] #these are the links active on time t
            
    for j in links:
        n1=j[0]
        n2=j[1]
        c=j[2]
        net.write(str(n1)+' '+str(n2)+' '+str(c)+'\n')
    net.close()
        
    t+=1
      
